@@ -382,7 +382,12 @@ def baseline_styler(typ):
 
 
 def show_experiments(
-    all_experiments, pager=True, no_timestamp=False, csv=False, **kwargs
+    all_experiments,
+    pager=True,
+    no_timestamp=False,
+    csv=False,
+    markdown=False,
+    **kwargs,
 ):
     from funcy.seqs import flatten as flatten_list
 
@@ -473,6 +478,7 @@ def show_experiments(
         header_styles=styles,
         row_styles=row_styles,
         csv=csv,
+        markdown=markdown,
     )
 
 
@@ -530,6 +536,7 @@ class CmdExperimentsShow(CmdBase):
                 iso=iso,
                 pager=not self.args.no_pager,
                 csv=self.args.csv,
+                markdown=self.args.markdown,
             )
         return 0
 
@@ -579,6 +586,8 @@ class CmdExperimentsDiff(CmdBase):
                     old=self.args.old,
                     on_empty_diff="diff not supported",
                     precision=precision if key == "metrics" else None,
+                    a_rev=self.args.a_rev,
+                    b_rev=self.args.b_rev,
                 )
 
         return 0
@@ -956,7 +965,10 @@ def add_parser(subparsers, parent_parser):
     )
     experiments_show_parser.add_argument(
         "--sort-order",
-        help="Sort order to use with --sort-by.",
+        help=(
+            "Sort order to use with --sort-by."
+            " Defaults to ascending ('asc')."
+        ),
         choices=("asc", "desc"),
         default="asc",
     )
@@ -986,6 +998,14 @@ def add_parser(subparsers, parent_parser):
         default=False,
         help="Print output in csv format instead of a human-readable table.",
     )
+    experiments_show_parser.add_argument(
+        "--md",
+        "--show-md",
+        action="store_true",
+        default=False,
+        dest="markdown",
+        help="Show tabulated output in the Markdown format (GFM).",
+    )
     experiments_show_parser.add_argument(
         "--precision",
         type=int,
@@ -1067,7 +1087,7 @@ def add_parser(subparsers, parent_parser):
         "--old",
         action="store_true",
         default=False,
-        help="Show old metric/param value.",
+        help="Show old ('a_rev') metric/param value.",
     )
     experiments_diff_parser.add_argument(
         "--no-path",
